import os
from pathlib import Path
from typing import Union
from .serializers import OrJson, Yaml
from .funcs import doesTextMatch
from .types import *




def hasEnv(name: str) -> bool: return os.getenv(name, None) is not None

def getEnv(): return {k:v for k,v in os.environ.items()}

def envToBool(name: str, default: str = 'false'): return bool(os.getenv(name, default).lower() in {'true', '1', 'yes'})

def envToType(value: str = None, valType = str):
    if value is None or value in {'none', 'None'}: return None
    return valType(value)

def envToList(name: str, sep: str = ',', default: List[str] = []):
    d = os.getenv(name, None)
    if d: return d.split(sep)
    return default

def envToDict(name: str, sep: str = '=', default: Dict[str, Any] = {}):
    d = envToList(name)
    if not d: return default
    rez = {}
    for i in d:
        for k,v in i.split(sep, 1):
            rez[k] = v
    return rez

def envToFloat(name: str, default: float = 0.0):
    d = os.getenv(name, None)
    if d is None: return default
    return envToType(d, float)

def envToInt(name: str, default: int = 0):
    d = os.getenv(name, None)
    if d is None: return default
    return envToType(d, int)

def envInVals(name: str, checkVals: List[str] = [], exact: bool = False, **kwargs) -> bool:
    d = os.getenv(name, None)
    if d is None: return False
    return doesTextMatch(text=d, items=checkVals, exact=exact, **kwargs)

def envToStr(name: str, default: str = ''):
    d = os.getenv(name, None)
    if d is None: return default
    return d


def toEnv(name: str, value: Any, override: bool = False):
    if not hasEnv(name) or override: os.environ[name] = str(value)


def toPath(path: Union[str, Path], resolve: bool = True) -> Path:
    if isinstance(path, str): path = Path(path)
    if resolve: path.resolve()
    return path


_LoadedEnvFiles = set()

def loadEnvFile(path: Union[str, Path], override: bool = False):
    global _LoadedEnvFiles
    path = toPath(path, resolve=True)
    if not path.exists(): return None
    if path.name in _LoadedEnvFiles: return True
    loader = None
    if path.suffix == '.json': loader = OrJson.loads
    elif path.suffix == '.yaml': loader = Yaml.loads
    if loader is None: return False
    data = loader(path.read_text())
    for k,v in data.items():
        toEnv(name=k, value=v, override=override)
    _LoadedEnvFiles.add(path.name)
    return True


__all__ = [
    'hasEnv',
    'getEnv',
    'envToBool',
    'envToType',
    'envToList',
    'envToDict',
    'envToFloat',
    'envToInt',
    'envInVals',
    'envToStr'
    'toEnv',
    'toPath',
    'loadEnvFile'
]


from pydantic import BaseModel, Field
from .types import *
from .serializers import OrJson
from .utils import to_camelcase


class BaseCls(BaseModel):
    class Config:
        arbitrary_types_allowed = True
    
    def get(self, name, default: Any = None):
        return getattr(self, name, default)

class BaseLazy(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = 'allow'
        alias_generator = to_camelcase
        json_loads = OrJson.loads
        json_dumps = OrJson.dumps

    def get(self, name, default: Any = None):
        return getattr(self, name, default)

class Validator(BaseCls):
    text: str
    include: List[str] = []
    exclude: List[str] = []
    exact: bool = False

class ValidatorArgs(BaseCls):
    include: List[str] = []
    exclude: List[str] = []

__all__ = [
    'BaseModel',
    'Field',
    'BaseCls'
    'BaseLazy'
]


from typing import Optional

class _classproperty(property):
    """
    The use this class as a decorator for your class property.
    Example:
        @classproperty
        def prop(cls):
            return "value"
    """
    def __get__(self, *args, **_) -> object:
        """
        This method gets called when a property value is requested.
        @return: The value of the property.
        """
        # apply the __get__ on the class, which is the second argument
        return super(_classproperty, self).__get__(args[1])

    def __set__(self, cls_or_instance, value: object) -> None:
        """
        This method gets called when a property value should be set.
        @param cls_or_instance: The class or instance of which the property should be changed.
        @param value: The new value.
        """
        # call this method only on the class, not the instance
        super(_classproperty, self).__set__(self.__get_class(cls_or_instance), value)

    def __delete__(self, cls_or_instance) -> None:
        """
        This method gets called when a property should be deleted.
        @param cls_or_instance: The class or instance of which the property should be deleted.
        """
        # call this method only on the class, not the instance
        super(_classproperty, self).__delete__(self.__get_class(cls_or_instance))

    def __get_class(self, cls_or_instance) -> type:
        """
        Get the class of an object if one is provided.
        @param cls_or_instance: Either an object or a class.
        @return: The class of the object or just the class again.
        """
        if isinstance(cls_or_instance, type): return cls_or_instance
        return type(cls_or_instance)

class _ClasspropertyMeta(type):
    """
    The class that uses the classproperty decorator must use this meta class if the setting and deleting of a class
    property should be supported.
    """
    def __setattr__(self, name: str, value: object) -> None:
        """
        Override of __setattr__ method to allow a classproperty.setter.
        @param name: The name of the attribute that should get a new value.
        @param value: The new value of the attribute.
        """
        cp_obj: Optional[_classproperty] = self.__get_classproperty_attr(name)
        if cp_obj: cp_obj.__set__(self, value)
        else: super(_ClasspropertyMeta, self).__setattr__(name, value)

    def __delattr__(self, name: str):
        """
        Override of __delattr__ method to allow a classproperty.deleter.
        @param str: The name of the attribute to delete.
        """
        cp_obj: Optional[_classproperty] = self.__get_classproperty_attr(name)
        if cp_obj: cp_obj.__delete__(self)
        else: super(_ClasspropertyMeta, self).__delattr__(name)

    def __get_classproperty_attr(self, name: str) -> Optional[_classproperty]:
        """
        Get a classproperty attribute from this class with the given name.
        @param name: The name of the attribute.
        @return: The classproperty object for the attribute. Or 'None' if it wasn't found.
        """
        # iterate through MRO list to get all attributes
        if (name in self.__dict__ and isinstance(self.__dict__[name], _classproperty)): return self.__dict__[name]
        else: return None


class classproperty(_classproperty):
    pass

class ClasspropertyMeta(_ClasspropertyMeta):
    pass

__all__ = [
    "classproperty", 
    "ClasspropertyMeta"
]



from . import types
from . import prop
from . import utils
from . import serializers
from . import models
from . import base
from . import funcs
from . import envs



from .prop import classproperty, ClasspropertyMeta

from .models import (
    BaseCls,
    BaseModel,
    BaseLazy,
    Field
)

from .base import (
    LazyCls,
    create_lazycls,
    clear_lazy_models,
    set_modulename
)

__all__ = [
    'classproperty',
    'ClasspropertyMeta',
    'BaseCls',
    'BaseModel',
    'BaseLazy',
    'Field',
    'LazyCls',
]


import orjson as _orjson
import yaml
from typing import Any

class OrJson:
    @classmethod
    def dumps(cls, obj, *args, default: Any = None, **kwargs):
        return _orjson.dumps(obj, default=default, *args, **kwargs).decode()

    @classmethod
    def loads(cls, obj, *args, **kwargs):
        return _orjson.loads(obj, *args, **kwargs)

class Yaml:
    @classmethod
    def dumps(cls, obj, *args, **kwargs):
        return yaml.dump(obj, *args, **kwargs)

    @classmethod
    def loads(cls, obj, *args, **kwargs):
        return yaml.load(obj, Loader=yaml.Loader, *args, **kwargs)

__all__ = [
    'OrJson',
    'Yaml'
]


import re
from datetime import datetime, timedelta
from functools import lru_cache, wraps
from pathlib import Path
from .types import *
from .models import ValidatorArgs

def getParentPath(p: str) -> Path: return Path(p).parent

def toList(text: Union[str, List[str]]):
    if isinstance(text, list): return text
    return [i for i in text.split(',') if i]


def setToMany(value: AnyMany) -> List[Any]:
    if not isinstance(value, list): value = [value]
    return value


def convertLog(*msgs, split_newline: bool = False):
    _msgs = []
    for msg in msgs:
        if isinstance(msg, list):
            for m in msg: _msgs.append(f'- {m}')
        elif isinstance(msg, dict):
            for k,v in msg.items():
                _msgs.append(f'- {k}: {v}')
        elif isinstance(msg, str):
            if split_newline:  _msgs.extend(msg.split('\n'))
            else: _msgs.append(msg)
        else: _msgs.append(f'{msg}')
    return [i for i in _msgs if i]

def doesTextValidate(text: str, include: List[str] = [], exclude: List[str] = [], exact: bool = False, **kwargs) -> bool:
    if not include and not exclude: return True
    _valid = False
    if exclude:
        for ex in exclude:
            if ((exact and ex == text)  or not exact and (ex in text or text in ex)): return False
        _valid = True
    if include:
        for inc in include:
            if ((exact and inc == text) or not exact and (inc in text or text in inc)): return True
    return _valid

def doesObjectListMatch(obj: List[Any], key: str = None, val: Any = None, exact: bool = False, **kwargs):
    if key and val and exact: return bool(len([i for i in obj if key in i and i.get(key) == val]) > 0)
    if key and exact: return bool(len([i for i in obj if key in i]) > 0)
    if val and exact: return bool(val in obj)
    if key and val: return bool(len([i for i in obj if key in i and (i.get(key) in val or val in i.get(key))]) > 0)
    if key: return bool(len([i for i in obj if (key in i or i in key)]) > 0)
    if val: return bool(val in obj or obj in val)
    return None

def doesObjectDictMatch(obj: DictAny, key: str = None, val: Any = None, exact: bool = False, allow_null: bool = False,  **kwargs):
    if key and val and exact: return bool(obj.get(key) == val)
    if key and exact and not allow_null: return bool(obj.get(key) is not None)
    if key and exact: return bool(str(obj.get(key, 'NULLVAL')) != 'NULLVAL')
    if val and exact: return bool(val in list(obj.values()))
    if key and val: return bool(obj.get(key) in val or val in obj.get(key))
    if key: return bool(obj.get(key))
    if val: return any(val in i or i in val for i in list(obj.values()))
    return None

def doesObjectStringMatch(obj: str, key: str = None, val: Any = None, exact: bool = False, **kwargs):
    if key and exact: return bool(obj == key)
    if val and exact: return bool(obj == val)
    if key: return bool(obj in key or key in obj)
    if val: return bool(obj in val or val in obj)
    return None


def doesObjectMatch(obj: Any, key: str = None, val: Any = None, exact: bool = False, allow_null: bool = False, **kwargs) -> bool:
    if not allow_null and obj is None: return False
    if isinstance(obj, list):
        rez = doesObjectListMatch(obj, key=key, val=val, exact=exact, **kwargs)
        if rez is not None: return rez
    if isinstance(obj, dict):
        rez = doesObjectDictMatch(obj, key=key, val=val, exact=exact, allow_null=allow_null, **kwargs)
        if rez is not None: return rez
    if isinstance(obj, str):
        rez = doesObjectStringMatch(obj, key=key, val=val, exact=exact, **kwargs)
        if rez is not None: return rez
    else:
        assert key is not None
        objval = getattr(obj, key, None)
        if val and exact and not allow_null: return bool(objval and objval == val)
        if not allow_null: return bool(objval is not None)
        if val: return bool(objval in val or val in objval)
    return False

def getObjectMatch(items: AnyMany, key: str, val: Any, returnMany: bool = False, exact: bool = False, allow_null: bool = False, **kwargs):
    items = setToMany(items)
    rez = []
    for i in items:
        if (doesObjectMatch(obj=i, key=key, val=val, exact=exact, allow_null=allow_null) and not returnMany): return i
        rez.append(i)
    return rez


def doesTextMatch(text: str, items: TextMany, exact: bool = False, valArgs: ValidatorArgs = None, **kwargs):
    items = setToMany(items)
    for i in items:
        if exact and i == text or (text in i or i in text):
            if valArgs: return doesTextValidate(i, exact=exact, **ValidatorArgs.dict())
            return True
    return False


def getTextMatch(text: str, items: TextMany, default: str = None, exact: bool = False, valArgs: ValidatorArgs = None, **kwargs):
    """Gets the First Matching Value from a List of String"""
    items = setToMany(items)
    for i in items:
        if exact and i == text or (text in i or i in text):
            if valArgs and doesTextValidate(i, exact=exact, **ValidatorArgs.dict()): return i
            return i
    return default

def getDictMatch(key: str, val: str, items: DictMany, default: DictAny = None, exact: bool = False, valArgs: ValidatorArgs = None, **kwargs):
    """Gets the First Matching Item from a Dict's Key = Value from a List of Dicts. Meant for Strings"""
    items = setToMany(items)
    for i in items:
        if exact and i.get(key, '') == val or (val in i.get(key, '') or i.get(key, '') in val):
            if valArgs and doesTextValidate(i[key], exact=exact, **ValidatorArgs.dict()): return i
            return i
    return default


def timedCache(seconds: int, maxsize: int = 128):
    def wrapper_cache(func):
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = datetime.utcnow() + func.lifetime
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if datetime.utcnow() >= func.expiration:
                func.cache_clear()
                func.expiration = datetime.utcnow() + func.lifetime
            return func(*args, **kwargs)
        return wrapped_func
    return wrapper_cache


def parseMetrics(data: Union[int, str]):
    if isinstance(data, int): return data
    if data.endswith('Ki'):
        return int(data.split('Ki')[0].strip()) * 1024
    if data.endswith('Mi'):
        return int(data.split('Mi')[0].strip()) * 1024**2
    if data.endswith('Gi'):
        return int(data.split('Gi')[0].strip()) * (1024**2 * 1024)
    if data.endswith('n'):
        return int(data.split('n')[0].strip())
    if data.endswith('m'):
        return int(data.split('m')[0].strip()) * 1000
    return int(data.strip())

def listToDict(data: List[Dict[str, str]], k: str, v: str):
    return {i[k]: i[v] for i in data}


def caseCamelToSnake(name):
    """
    Convert a name in camel case to snake case.
    Arguments:
    name -- The name to convert.
    Returns:
    The name in snake case.
    """
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


__all__ = [
    'toList',
    'listToDict',
    'parseMetrics',
    'timedCache',
    'convertLog',
    'getObjectMatch',
    'doesTextMatch',
    'getTextMatch',
    'getDictMatch',
    'doesTextValidate',
    'setToMany',
    'getParentPath',
    'caseCamelToSnake'
]


import os
import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info.major != 3:
    raise RuntimeError("This package requires Python 3+")

version = '0.0.2'
pkg_name = 'lazycls'
gitrepo = 'trisongz/lazycls'
root = Path(__file__).parent

requirements = [
    'pydantic',
    'orjson',
    'pyyaml'
]

args = {
    'packages': find_packages(include = ['lazycls', 'lazycls.*']),
    'install_requires': requirements,
    'long_description': root.joinpath('README.md').read_text(encoding='utf-8'),
    'entry_points': {}
}

setup(
    name=pkg_name,
    version=version,
    url=f'https://github.com/{gitrepo}',
    license='MIT Style',
    description='Dynamic Dataclasses for the Super Lazy',
    author='Tri Songz',
    author_email='ts@growthengineai.com',
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
    ],
    **args
)